/*
    This file is part of the 'yarec' package.
    Copyright (C) 1998  Ralf W. Stephan <stephan@tmt.de>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "soundHandler.h"

#ifdef SOUND_LINUX

#include <unistd.h>
#include <stdio.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/soundcard.h>
#include <linux/types.h>
#include "ui/uiHandler.h"
#include "file/file.h"

#define BUFSIZE 65536

static int sd_, opened_, play_f_, r_, b_, c_, frag_, blocksize_;

void LinuxSound_init (int *r, int *b, int *c, int play_f)
{
  int t, ret;
  static char buf[256];

  if (opened_) return;

  sd_ = open ("/dev/dsp", play_f? O_WRONLY:O_RDONLY);
  if (sd_ < 0)
  	Exception_fatal (_("Could not open /dev/dsp.  No sound driver on your system?"));

  t = *b;
  ret = ioctl (sd_, SNDCTL_DSP_SAMPLESIZE, &t);
  if (ret == -1)
  	Exception_fatal ("SNDCTL_DSP_SAMPLESIZE");
  if (t != *b) {
  	if (play_f) {
  		sprintf (buf, _("Tried to play %d-bit sample but couldn't"), *b);
  		Exception_fatal (buf);
  	}
  	sprintf (buf, _("Sample size %d bits not supported by driver.\nUsing %d bits."), *b, t);
  	UIHandler_msg (buf);
  	*b = t;
  }
  b_ = t;

  t = *c;
  ret = ioctl (sd_, SNDCTL_DSP_CHANNELS, &t);
  if (ret == -1)
  	Exception_fatal ("SNDCTL_DSP_CHANNELS");
  if (t != *c) {
  	if (play_f) {
  		sprintf (buf, _("Tried to play %d-channel sample but couldn't"), *c);
  		Exception_fatal (buf);
  	}
  	sprintf (buf, _("%d channel/s not supported by driver.\nUsing %d channel/s."), *c, t);
  	UIHandler_msg (buf);
  	*c = t;
  }
  c_ = t;
  
  t = *r;
  ret = ioctl (sd_, SNDCTL_DSP_SPEED, &t);
  if (ret == -1)
  	Exception_fatal ("SNDCTL_DSP_SPEED");
  if (t != *r) {
  	sprintf (buf, _("Sampling rate %d Hz not supported by driver.\nUsing %d Hz."), *r, t);
  	UIHandler_msg (buf);
  	*r = t;
  }
  r_ = t;

  /* this is stolen from krecord/oss.cpp */
  blocksize_ = 50*c_*r_/1000;
  for (frag_ = 0; blocksize_ != 1; frag_++)
  	blocksize_ >>= 1;
  frag_ |= 0x7fff0000;
  if (-1 == ioctl (sd_, SNDCTL_DSP_SETFRAGMENT, &frag_))
        perror("ioctl SNDCTL_DSP_SETFRAGMENT");
  if (-1 == ioctl (sd_, SNDCTL_DSP_GETBLKSIZE,  &blocksize_))
         perror("ioctl SNDCTL_DSP_GETBLKSIZE"); 
  
  if (!play_f) {
	t = ~PCM_ENABLE_INPUT;
	ioctl(sd_,SNDCTL_DSP_SETTRIGGER,&t);
	t = PCM_ENABLE_INPUT;
	ioctl(sd_,SNDCTL_DSP_SETTRIGGER,&t);
	}

  opened_ = 1;
  play_f_ = play_f;
}

void LinuxSound_open_again()
{
  int t, ret;

  if (opened_) return;

  sd_ = open ("/dev/dsp", play_f_? O_WRONLY:O_RDONLY);
  if (sd_ < 0)
  	UIHandler_msg (_("Could not reopen /dev/dsp"));

  t = b_;
  ret = ioctl (sd_, SNDCTL_DSP_SAMPLESIZE, &t);

  t = c_;
  ret = ioctl (sd_, SNDCTL_DSP_CHANNELS, &t);
  
  t = r_;
  ret = ioctl (sd_, SNDCTL_DSP_SPEED, &t);
  
  t = frag_;
  ret = ioctl (sd_, SNDCTL_DSP_SETFRAGMENT, &t);

  t = blocksize_;
  ret = ioctl (sd_, SNDCTL_DSP_GETBLKSIZE,  &t);

  if (!play_f_) {
	t = ~PCM_ENABLE_INPUT;
	ioctl(sd_,SNDCTL_DSP_SETTRIGGER,&t);
	t = PCM_ENABLE_INPUT;
	ioctl(sd_,SNDCTL_DSP_SETTRIGGER,&t);
	}

 opened_ = 1;
}

long LinuxSound_play_chunk (u32 *p, long size)
{
  if (!opened_ || !play_f_)
  	return 0;

  size = write (sd_, (char*)p, size * (b_/8));
  return size/(b_/8);
}

long LinuxSound_rec_chunk (u32 *p, long size)
{
  static u8 buf8 [BUFSIZE];
  static u16 buf16 [BUFSIZE];
  u16* p16 = buf16;
  u8* p8 = buf8;
  int i;

  if (!opened_ || play_f_)
  	return 0;

  if (size > BUFSIZE)
  	Exception_fatal ("recbufsize too small!");
  
  switch (b_) {
  case  8: size = read (sd_, buf8, size);
           for (i=0; i<size; i++)
           	*p++ = ((*p8++ ^ 0x80) << 24);
           break;
  case 16: size = read (sd_, buf16, size*2);
           size /= 2;
           for (i=0; i<size; i++) {
           	*p++ = (*p16++ << 16);
           }
           break;
  case 32: size = read (sd_, p, size*4);
           size /= 4;
           break;
  default: Exception_fatal ("Riot in LinuxSound_rec_chunk()!");
  }
  return size;
}

void LinuxSound_tmp_close()
{
  if (sd_ > 0)
  	close (sd_);
  opened_ = 0;
}

void LinuxSound_close()
{
  LinuxSound_tmp_close();
}

#endif
